#include "tsdb.h"

#include <clickhouse/client.h>
#include <clickhouse/columns/string.h>
#include <fmt/format.h>
#include <spdlog/spdlog.h>

#include "shared.hpp"

namespace tsdb_cpp {

// Forward declaration
void tsdb_entry_notify_batch_close(tsdb_entry_impl* entry);
std::string_view tsdb_entry_get_correlation_id(tsdb_entry_impl* entry) noexcept;
clickhouse::Client& tsdb_entry_get_client(tsdb_entry_impl* entry);

struct tsdb_batch_insertion::impl {
    impl(tsdb_entry_impl* parent, std::string_view table_name, std::string_view metric_name)
        : parent_(parent),
          table_name_(table_name),
          metric_name_(metric_name),
          col_metric_name_(std::make_shared<clickhouse::ColumnString>()),
          col_point_name_(std::make_shared<clickhouse::ColumnString>()),
          col_value_(std::make_shared<clickhouse::ColumnFloat64>()),
          col_nanoseconds_(std::make_shared<clickhouse::ColumnInt64>()) {
        correlation_id_ = tsdb_shared::random_uuid();
        spdlog::debug("TSDB-lf batch created from entry {}, correlation ID: {}", tsdb_entry_get_correlation_id(parent_), correlation_id_);
    }
    ~impl() {
        spdlog::debug("TSDB-lf batch closed, correlation ID: {}", correlation_id_);
        // Notify the parent that the batch is closed
        tsdb_entry_notify_batch_close(parent_);
    }

    void add_point(std::string_view name, double value, long long nanoseconds) {
        // spdlog::trace("TSDB-lf batch point added, correlation ID: {}", correlation_id_);
        spdlog::trace("TSDB-lf batch point added, cur count = {}, correlation ID: {}", col_metric_name_->Size() + 1, correlation_id_);
        col_metric_name_->Append(metric_name_);
        col_point_name_->Append(name);
        col_value_->Append(value);
        col_nanoseconds_->Append(nanoseconds);
    }

    void commit() {
        spdlog::debug("TSDB-lf batch committing, correlation ID: {}", correlation_id_);
        clickhouse::Block block;
        block.AppendColumn("metric_name", col_metric_name_);
        block.AppendColumn("point_name", col_point_name_);
        block.AppendColumn("value", col_value_);
        block.AppendColumn("nanoseconds", col_nanoseconds_);
        tsdb_entry_get_client(parent_).Insert(table_name_, block);
        spdlog::debug("TSDB-lf batch committed, correlation ID: {}", correlation_id_);
    }

private:
    tsdb_entry_impl* const parent_;
    std::string table_name_;
    std::string metric_name_;
    std::shared_ptr<clickhouse::ColumnString> col_metric_name_;
    std::shared_ptr<clickhouse::ColumnString> col_point_name_;
    std::shared_ptr<clickhouse::ColumnFloat64> col_value_;
    std::shared_ptr<clickhouse::ColumnInt64> col_nanoseconds_;
    std::string correlation_id_;
};

tsdb_batch_insertion::tsdb_batch_insertion(std::unique_ptr<impl> impl) : impl_(std::move(impl)) {
}
tsdb_batch_insertion::~tsdb_batch_insertion() {
    close();
}
tsdb_batch_insertion::tsdb_batch_insertion(tsdb_batch_insertion&& other) : impl_(std::move(other.impl_)) {
}
void tsdb_batch_insertion::close() {
    if (auto impl = std::move(impl_)) {
        impl->commit();
    }
}
void tsdb_batch_insertion::add_point(std::string_view name, double value, long long nanoseconds) {
    impl_->add_point(name, value, nanoseconds);
}

struct tsdb_entry_impl {
    tsdb_entry_impl(std::string_view remote, uint16_t port) : client_(clickhouse::ClientOptions().SetHost(remote.data()).SetPort(port)) {
        // Initialize the TSDB environment
        tsdb_shared::init_tsdb_env();

        correlation_id_ = tsdb_shared::random_uuid();
        spdlog::debug("TSDB-lf entry created, correlation ID: {}", correlation_id_);
    }
    ~tsdb_entry_impl() { spdlog::debug("TSDB-lf entry closed, correlation ID: {}", correlation_id_); }

    void close() {
        if (open_batch_count_ > 0) {
            // Fail-fast
            throw std::runtime_error("Cannot close tsdb_entry while there are open batches");
        }
    }

    void create_table_inner(std::string_view table_name, std::string_view engine, bool can_order, std::string_view extras = "") {
        table_name_ = table_name;
        auto query  = fmt::format(
            "CREATE TABLE IF NOT EXISTS {} ("
             "metric_name LowCardinality(String) COMMENT 'Metric name',"
             "point_name LowCardinality(String) COMMENT 'Point name',"
             "tags Map(LowCardinality(String), LowCardinality(String)) COMMENT 'Point tags',"
             "value Float64 COMMENT 'Point value',"
             "nanoseconds Int64 COMMENT 'Point time in nanoseconds' Codec(DoubleDelta, LZ4)"
             ") ENGINE = {}",
            table_name, engine);
        if (can_order) {
            query += " ORDER BY (metric_name, point_name, nanoseconds)";
        }
        if (!extras.empty()) {
            query += " ";
            query += extras;
        }
        client_.Execute(query);
    }

    void drop_table_inner(std::string_view table_name) {
        auto query = fmt::format("DROP TABLE IF EXISTS {}", table_name);
        client_.Execute(query);
    }

    void set_active_table(std::string_view table_name) { table_name_ = table_name; }

    void set_metric_name(std::string_view metric_name) { metric_name_ = metric_name; }

    void insert_point(const point& p) { insert_point(p.name_, p.value_, p.nanoseconds_); }
    void insert_point(std::string_view name, double value, long long nanoseconds) {
        spdlog::debug("TSDB-lf point inserting, correlation ID: {}", correlation_id_);
        clickhouse::Block block;
        auto col1 = std::make_shared<clickhouse::ColumnString>();
        col1->Append(metric_name_);
        block.AppendColumn("metric_name", col1);
        auto col2 = std::make_shared<clickhouse::ColumnString>();
        col2->Append(name);
        block.AppendColumn("point_name", col2);
        auto col3 = std::make_shared<clickhouse::ColumnFloat64>();
        col3->Append(value);
        block.AppendColumn("value", col3);
        auto col4 = std::make_shared<clickhouse::ColumnInt64>();
        col4->Append(nanoseconds);
        block.AppendColumn("nanoseconds", col4);
        client_.Insert(table_name_, block);
        spdlog::debug("TSDB-lf point inserted, correlation ID: {}", correlation_id_);
    }

    void insert_points(const std::vector<point>& points) {
        spdlog::debug("TSDB-lf [!LEGACY!] points inserting, correlation ID: {}", correlation_id_);
        auto batch = new_batch_insert();
        for (const auto& p : points) {
            batch.add_point(p.name_, p.value_, p.nanoseconds_);
        }
        batch.close();
        spdlog::debug("TSDB-lf [!LEGACY!] points inserted, correlation ID: {}", correlation_id_);
    }

    tsdb_batch_insertion new_batch_insert() {
        auto batch = tsdb_batch_insertion(std::make_unique<tsdb_batch_insertion::impl>(this, table_name_, metric_name_));
        open_batch_count_++;
        return batch;
    }

    friend void tsdb_entry_notify_batch_close(tsdb_entry_impl* entry) {
        if (entry->open_batch_count_ == 0) {
            // Fail-fast
            throw std::runtime_error("Cannot close tsdb_entry while there are open batches");
        }
        entry->open_batch_count_--;
    }
    friend std::string_view tsdb_entry_get_correlation_id(tsdb_entry_impl* entry) noexcept { return entry->correlation_id_; }
    friend clickhouse::Client& tsdb_entry_get_client(tsdb_entry_impl* entry) { return entry->client_; }

private:
    clickhouse::Client client_;
    std::string table_name_;
    std::string metric_name_ = "";
    size_t open_batch_count_ = 0;
    std::string correlation_id_;
};

tsdb_entry::tsdb_entry(std::string_view remote, uint16_t port) : impl_(std::make_unique<tsdb_entry_impl>(remote, port)) {
}
tsdb_entry::~tsdb_entry() {
    close();
}
tsdb_entry::tsdb_entry(tsdb_entry&& other) : impl_(std::move(other.impl_)) {
}
void tsdb_entry::close() {
    if (auto impl = std::move(impl_)) {
        impl->close();
    }
}
void tsdb_entry::create_table() {
    impl_->create_table_inner("tsdb_cpp", "MergeTree()", true);
}
void tsdb_entry::create_memory_table() {
    impl_->create_table_inner("tsdb_cpp_memory", "Memory()", false);
}
void tsdb_entry::drop_table() {
    impl_->drop_table_inner("tsdb_cpp");
}
void tsdb_entry::drop_memory_table() {
    impl_->drop_table_inner("tsdb_cpp_memory");
}
void tsdb_entry::set_active_table(std::string_view table_name) {
    impl_->set_active_table(table_name);
}
void tsdb_entry::set_metric_name(std::string_view metric_name) {
    impl_->set_metric_name(metric_name);
}
void tsdb_entry::insert_point(const point& p) {
    impl_->insert_point(p);
}
void tsdb_entry::insert_points(const std::vector<point>& points) {
    impl_->insert_points(points);
}
void tsdb_entry::insert_point(std::string_view name, double value, long long nanoseconds) {
    impl_->insert_point(name, value, nanoseconds);
}
tsdb_batch_insertion tsdb_entry::new_batch_insert() {
    return impl_->new_batch_insert();
}

}  // namespace tsdb_cpp
